/*
	Miranda Installer - Installs nightlies and Miranda addons.
	Copyright (C) 2002-2003 Goblineye Entertainment

	Authors: Saar (Tornado) and Kai (kai_b)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#pragma once

#include "MirandaInstaller.h"
#pragma hdrstop

// these aren't flags, they're unique
// installation types associated with the lower 8 bits of each
// installation item
#define INSTALLTYPE_PLUGIN 0
#define INSTALLTYPE_ICON 1
#define INSTALLTYPE_SOUND 2
#define INSTALLTYPE_LANGUAGE 3
#define INSTALLTYPE_TOOL 4
#define INSTALLTYPE_SOURCE 5
#define INSTALLTYPE_DOC 6
#define INSTALLTYPE_SKIN 7
#define INSTALLTYPE_NIGHTLY 8

#define INSTALLSCRIPT_FILENAME "InstallScript.xml" // case-insensitive

// these are msgs sent to dialog
// sent by extraction routine and worker thread
// originally used in Wassup Updater
#define MIRINST_INFOMSG_STARTING (WM_USER + 20) // starting a new batch
#define MIRINST_INFOMSG_STARTITEM (WM_USER + 21) // wParam is ID (from ListView) - starting a new installation
#define MIRINST_INFOMSG_FINISHEDITEM (WM_USER + 22) // wParam is ID, lParam is 2 if we canceled, lParam is 0 if everything went well, 1 if error
#define MIRINST_INFOMSG_EXTRACTING (WM_USER + 23) // extracting a file. wParam is package ID. lParam is a pointer to the file name - extraction will begin in a sec
#define MIRINST_INFOMSG_DONE (WM_USER + 24) // the thread finished work. yippie. wParam is true if error occured (no need to pop msg)

// declared in Worker.h
// this is just so the compiler will keep quiet
typedef struct GENERAL_FILELIST_TYPE GENERAL_FILELIST;


typedef struct ADDON_BASICINFO_TYPE
{
	char *lpszName; // name (taken from script, if not found in script, filename is used)
	char *lpszAuthor; // author
	char *lpszVersion;
	char *lpszType; // priority first goes to type declared in script, otherwise - extension is tested.
	bool fScripted; // is an install script available?
	WORD wTotalPackages; // if this info is not available, this will be 0
	WORD wMainPackages;

	// these are used internally, in a sense
	bool fGotEmptyPackage; // parser found an empty package (i.e. a package with no files)
	GENERAL_FILELIST *pFileList; // pFileList lists ALL files mentioned by the script! everything!! :)
	// it's up to the caller to go through the list and find irregularities like dupes, and invalid files
	// slashes are reversed before being added to the list (to slashes - '/')
	// btw: nameless packages are not deal breakers, a stupid name is used for them instead
	// deal breakers: (w/o - this package cannot be installed)
	// if ((lpszType == NULL) || (lpszName == NULL)) -> no given type (scripted) or by extension
	// if ((fScripted) && (wMainPackages == 0)) -> install script is no good, an error should be given
	// if (fGotEmptyPackage) -> an empty package was found
	// if (after_going_thru_file_list,dupes_and_invalid_files_were_found) -> :)
}	ADDON_BASICINFO;

// each listview item points to a ADDON_LISTINFO structure. must be deallocated later
typedef struct ADDON_LISTINFO_TYPE
{
	byte bType; // type. (INSTALLTYPE_*)
	char *lpFilePath; // absolute path to file
}	ADDON_LISTINFO;

// defining it in the header gives way too many problems
extern const char* INSTALLATION_EXTS_ARR[10];

void InstallList_AddFile(HWND hwndDlg,const char *lpFilePath,bool fQuiet=false); // install a file. :) this function gives out errors and such on its own (fQuiet means that msg boxes won't be shown on invalid packages. good for adding a bunch of files at the same time)
void InstallList_AddDirectory(HWND hwndDlg,const char *lpDirectoryPath);